"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ReadableStreamController = void 0;
const outvariant_1 = require("outvariant");
const ReadableStream_1 = require("./ReadableStream");
async function noOpAlgorithm() {
    return;
}
class ReadableStreamController {
    state;
    stream;
    startAlgorithm;
    pullAlgorithm;
    cancelAlgorithm;
    sizeAlgorithm;
    queue;
    queueTotalSize;
    pendingPullIntos;
    constructor(options) {
        this.stream = options.stream;
        this.state = {
            started: false,
            pulling: false,
            pullAgain: false,
            closeRequested: false,
        };
        this.startAlgorithm = options.startAlgorithm;
        this.pullAlgorithm = options.pullAlgorithm || noOpAlgorithm;
        this.cancelAlgorithm = options.cancelAlgorithm || noOpAlgorithm;
        this.sizeAlgorithm = options.sizeAlgorithm;
        this.queue = [];
        this.queueTotalSize = 0;
        this.pendingPullIntos = [];
        /**
         * @see https://github.com/nodejs/node/blob/561f7fe941929d6c10b82b8250c04afb0693e4f3/lib/internal/webstreams/readablestream.js#L1970-L1978
         */
        this.startAlgorithm?.(this)
            .then(() => {
            this.state.started = true;
            this.pullIfNeeded();
        })
            .catch(this.triggerError);
    }
    /**
     * @see https://github.com/nodejs/node/blob/561f7fe941929d6c10b82b8250c04afb0693e4f3/lib/internal/webstreams/readablestream.js#L1924
     */
    pull(readRequest) {
        if (this.queue?.length) {
            const chunk = this.dequeueChunk();
            if (this.state.closeRequested && !this.queue.length) {
                this.clearAlgorithms();
                this.close();
            }
            else {
                this.pullIfNeeded();
            }
            readRequest.chunk(chunk);
            return;
        }
        const reader = this.stream['reader'];
        (0, outvariant_1.invariant)(reader, 'Reader not set');
        reader.readRequests.push(readRequest);
        this.pullIfNeeded();
    }
    getReadRequestCount() {
        return this.stream['reader']?.readRequests.length ?? 0;
    }
    dequeueChunk() {
        const chunk = this.queue.shift();
        (0, outvariant_1.invariant)(chunk, 'Failed to dequeue a chunk');
        this.queueTotalSize = Math.max(0, this.queueTotalSize - chunk.size);
        return chunk.value;
    }
    /** @see https://github.com/nodejs/node/blob/561f7fe941929d6c10b82b8250c04afb0693e4f3/lib/internal/webstreams/readablestream.js#L1804 */
    enqueue(chunk) {
        // If the stream is locked and is being read, then push the queued chunk
        // directly to the read request.
        if (this.stream.locked && this.getReadRequestCount()) {
            const readRequest = this.stream['reader']?.readRequests.shift();
            if (readRequest) {
                readRequest.chunk(chunk);
            }
        }
        else {
            try {
                const chunkSize = this.sizeAlgorithm(chunk);
                this.enqueueValueWithSize(chunk, chunkSize);
            }
            catch (error) {
                if (error instanceof Error) {
                    this.triggerError(error);
                }
                throw error;
            }
        }
        this.pullIfNeeded();
    }
    /** @see https://github.com/nodejs/node/blob/561f7fe941929d6c10b82b8250c04afb0693e4f3/lib/internal/webstreams/util.js#L156 */
    enqueueValueWithSize(chunk, chunkSize) {
        this.queue.push({
            value: chunk,
            size: chunkSize,
        });
        this.queueTotalSize += chunkSize;
    }
    /**
     * @see https://github.com/nodejs/node/blob/561f7fe941929d6c10b82b8250c04afb0693e4f3/lib/internal/webstreams/readablestream.js#L1794
     */
    close() {
        this.state.closeRequested = true;
        if (!this.queue.length) {
            this.clearAlgorithms();
            this.stream.close();
        }
    }
    async cancel(reason) {
        this.clearQueue();
        try {
            const cancelResult = this.cancelAlgorithm?.(reason);
            return cancelResult;
        }
        finally {
            this.clearAlgorithms();
        }
    }
    canCloseOrEnqueue() {
        return (this.state.closeRequested &&
            this.stream['state'] === ReadableStream_1.ReadableStreamState.Readable);
    }
    /**
     * @see https://github.com/nodejs/node/blob/561f7fe941929d6c10b82b8250c04afb0693e4f3/lib/internal/webstreams/readablestream.js#L1876
     */
    pullIfNeeded() {
        if (!this.shouldCallPull()) {
            return;
        }
        if (this.state.pulling) {
            this.state.pullAgain = true;
            return;
        }
        this.state.pulling = true;
        this.pullAlgorithm?.()
            .then(() => {
            this.state.pulling = false;
            if (this.state.pullAgain) {
                this.state.pullAgain = false;
                this.pullIfNeeded();
            }
        })
            .catch((error) => this.triggerError(error));
    }
    shouldCallPull() {
        if (!this.canCloseOrEnqueue() || !this.state.started) {
            return false;
        }
        if (this.stream.locked && this.getReadRequestCount()) {
            return true;
        }
        const desizedSize = 0;
        (0, outvariant_1.invariant)(desizedSize !== null, 'Failed to check controller pull: desized size is null');
        return desizedSize > 0;
    }
    triggerError(error) {
        this.clearQueue();
        this.clearAlgorithms();
        this.stream['triggerError'](error);
    }
    clearQueue() {
        this.queue = [];
        this.queueTotalSize = 0;
    }
    clearAlgorithms() {
        this.pullAlgorithm = undefined;
        this.cancelAlgorithm = undefined;
        this.sizeAlgorithm = undefined;
    }
}
exports.ReadableStreamController = ReadableStreamController;
//# sourceMappingURL=ReadableStreamController.js.map